package ec.app.cosolvable;

import ec.EvolutionState;
import ec.Individual;
import ec.app.parity.Parity;
import ec.gp.GPIndividual;
import ec.gp.koza.KozaFitness;
import ec.gp.semantic.GPSemantic;
import ec.gp.semantic.GPSemanticIndividual;
import ec.simple.SimpleFitness;
import ec.util.Parameter;

public class CosolvableComparator extends Parity {

	
	public static final String P_COSOLVE = "cosolve";
	public static final String P_TYPE = "type";
	public static final String P_FIT = "fit";
	public static final String P_SIZE = "size";
	public static final String P_PROB = "prob";
	
	public static final String COSOLVE_IFS = "ifs";
	public static final String COSOLVE_IFS2 = "ifs2";	
	public static final String COSOLVE_NONE = "none";
	
	public static final String COSOLVE_PRIM = "prim";
	public static final String COSOLVE_TRIO = "trio";
	
	public static final String COSOLVE_CMP = "cmp";
	public static final String COSOLVE_EQU = "equ";
	
	public String prob = null;
	public String fit = null;
	public String type = null;
	public boolean cosolve = false;
	
	private double S [][];
	private int halfSize;

	
	public void setup(final EvolutionState state, final Parameter base) {
		super.setup(state,base);
		
		prob = state.parameters.getString(base.push(P_PROB),null);
		if ( !(prob.equals(COSOLVE_CMP) || prob.equals(COSOLVE_EQU) ) ) {
			state.output.fatal("Wrong comparator problem", base.push(P_PROB));
		}
		
		fit = state.parameters.getString(base.push(P_FIT),null);
		if ( !(fit.equals(COSOLVE_PRIM) || fit.equals(COSOLVE_TRIO) ) ) {
			state.output.fatal("Wrong cosolve fit", base.push(P_FIT));
		}
				
		type = state.parameters.getString(base.push(P_TYPE),null);
		if ( !( type.equals(COSOLVE_IFS) || type.equals(COSOLVE_IFS2) || type.equals(COSOLVE_NONE) ) ) {
			state.output.fatal("Wrong cosolve routine", base.push(P_TYPE));
		}					
		
		cosolve = state.parameters.getBoolean(base.push(P_COSOLVE), null, false);
		
		int n = state.parameters.getInt(base.push(P_SIZE), null); 
		
    	S = new double[ n ][ n ];

    	halfSize = 1;
        for(int x=0;x<numBits/2;x++)
            halfSize *=2;   // safer than Math.pow()
     			
		state.output.message("cosolve " + type + " is " + (cosolve==true?"ON":"OFF") );
	}

		
	public void evaluate(final EvolutionState state, 
	        final Individual ind, 
	        final int subpopulation,
	        final int threadnum)
	        {
	        if (!ind.evaluated)  // don't bother reevaluating
	            {
	            int sum = 0;
	                
	            for(bits=0;bits<totalSize;bits++)
	                {
	                int tb = 0;

	                int num1 = bits / (halfSize);
	                int num2 = bits % (halfSize);
	                
	                if (prob.equals(COSOLVE_EQU)) {
	                	if (num1 == num2) tb = 1;
	                } else {
	                	if (num1 < num2) tb = 1;
	                }
	               	                               
	                ((GPIndividual)ind).trees[0].child.eval(
	                    state,threadnum,input,stack,((GPIndividual)ind),this);

	                ((GPSemanticIndividual)ind).semantic.semantic[ bits ]  =  0;
	                
	                if (input.x == tb) {
	                	sum++;
            			((GPSemanticIndividual)ind).semantic.semantic[ bits ]  =  1;
	                }
	                	                
	                }
	                	
	            // the fitness better be KozaFitness!
	            KozaFitness f = ((GPSemanticIndividual) ind).semantic.regressionFit;
	            f.setStandardizedFitness(state,(float)(totalSize - sum));
	            f.hits = sum;
	            ind.evaluated = true;
	            }
	        }

	
	public void finishEvaluating(final EvolutionState state, final int threadnum) {
    	
    	int n = ((GPSemanticIndividual) state.population.subpops[0].individuals[0]).semantic.semantic.length;    	
    	
    	for (int x = 0; x < n; ++x)
			for (int y = 0; y < n; ++y)
				S[x][y] = 0;
    	
    	for (int i = 0; i < state.population.subpops[0].individuals.length; ++i) {
    		GPSemanticIndividual ind = (GPSemanticIndividual) state.population.subpops[0].individuals[i];
    		for (int x = 0; x < n; ++x)
    			for (int y = 0; y < n; ++y)
    				S[x][y] += ((GPSemantic)(ind.semantic)).semantic[x] * ((GPSemantic)(ind.semantic)).semantic[y];     				
    	}
    	
    	int m = state.population.subpops[0].individuals.length;
    	for (int i = 0; i < m; ++i) {
    		GPSemanticIndividual ind = (GPSemanticIndividual) state.population.subpops[0].individuals[i];
    		double sum = 0;
    		
    		//prim: fit(I) = \sum_{i,j,S[i,j]!=0} I[i,j] * 1 / S[i,j]    		
    		//zij := si*sj/m - sij
    		//trio: fit(I) = \sum_{i,j,i>j} I[i,j] * 1 / (m/4 - zij)
    		    	    	    		
    		double [] I = ((GPSemantic)(ind.semantic)).semantic;

    		if (type.equals(COSOLVE_IFS))
	    		for (int x = 0; x < n; ++x) {
	    			int y = x;
	    			if (S[x][y] != 0)
	    				if (fit.equals(COSOLVE_PRIM)) {
	    					sum += I[x] * I[y] / S[x][y];
	    				} else {
	    					sum += I[x] * I[y] / (m/4 - S[x][x]*S[y][y]/m + S[x][y]);
	    				}
	    		}
    		
    		if (type.equals(COSOLVE_IFS2))
	    		for (int x = 0; x < n; ++x)
	    			for (int y = x + 1; y < n; ++y)
	    				if (S[x][y] != 0)
	    					if (fit.equals(COSOLVE_PRIM)) {
		    					sum += I[x] * I[y] / S[x][y];
		    				} else {
		    					sum += I[x] * I[y] / (m/4 - S[x][x]*S[y][y]/m + S[x][y]);
		    				}
    		    		
    		SimpleFitness f = ind.semantic.cosolvableFit;    		 
    		f.setFitness( state, (float)sum, false);
    		
    		if (cosolve) {
    			( (SimpleFitness) ind.fitness ).setFitness( state, ind.semantic.cosolvableFit.fitness(), false);
    		} else {
    			( (SimpleFitness) ind.fitness ).setFitness( state, ind.semantic.regressionFit.fitness(), false);
    		}
    		
    		ind.evaluated = true;    		
    	}
    }

	
	
	
}
