package ec.app.cosolvable;

import ec.EvolutionState;
import ec.Individual;
import ec.app.multiplexer.Fast;
import ec.app.multiplexer.Multiplexer;
import ec.app.multiplexer.MultiplexerData;
import ec.gp.GPIndividual;
import ec.gp.koza.KozaFitness;
import ec.gp.semantic.GPSemantic;
import ec.gp.semantic.GPSemanticIndividual;
import ec.simple.SimpleFitness;
import ec.util.Parameter;

public class CosolvableMultiplexer extends Multiplexer {

	public static final String P_COSOLVE = "cosolve";
	public static final String P_TYPE = "type";
	public static final String P_FIT = "fit";
	public static final String P_SIZE = "size";
		
	public static final String COSOLVE_IFS = "ifs";
	public static final String COSOLVE_IFS2 = "ifs2";
	public static final String COSOLVE_NONE = "none";
	
	public static final String COSOLVE_PRIM = "prim";
	public static final String COSOLVE_TRIO = "trio";
	
	public static final String COSOLVE_CMP = "cmp";
	public static final String COSOLVE_EQU = "equ";
		
	public String fit = null;
	public String type = null;
	public boolean cosolve = false;	
	
	public int totalSize;
	
	private double S [][];
	
	public void setup(final EvolutionState state, final Parameter base) {
		super.setup(state,base);
				
		fit = state.parameters.getString(base.push(P_FIT),null);
		if ( !(fit.equals(COSOLVE_PRIM) || fit.equals(COSOLVE_TRIO) ) ) {
			state.output.fatal("Wrong cosolve fit", base.push(P_FIT));
		}
				
		type = state.parameters.getString(base.push(P_TYPE),null);
		if ( !( type.equals(COSOLVE_IFS) || type.equals(COSOLVE_IFS2) || type.equals(COSOLVE_NONE) ) ) {
			state.output.fatal("Wrong cosolve routine", base.push(P_TYPE));
		}					
		
		cosolve = state.parameters.getBoolean(base.push(P_COSOLVE), null, false);
				
		if (bits==1)
			totalSize = Fast.M_3_SIZE;
        else if (bits==2)
			totalSize = Fast.M_6_SIZE;
        else // (bits==3)
			totalSize = Fast.M_11_SIZE;
	            		
    	S = new double[ totalSize ][ totalSize ];
    			
		state.output.message("cosolve " + type + " is " + (cosolve==true?"ON":"OFF") );
	}

	
	
	public void evaluate(final EvolutionState state, 
	        final Individual ind, 
	        final int subpopulation,
	        final int threadnum)
	        {
		
	     input.status = (byte)bits;

	        if (!ind.evaluated)  // don't bother reevaluating
	            {
	            int sum = 0;
	                
	            ((GPIndividual)ind).trees[0].child.eval(
	                state,threadnum,input,stack,((GPIndividual)ind),this);
	                
	            if (bits==1)
	                {
	                byte item1 = input.dat_3;
	                byte item2 = Fast.M_3[Fast.M_3_OUTPUT];
	                for(int y=0;y<MultiplexerData.MULTI_3_BITLENGTH;y++)
	                    {
	                    // if the first bit matches, grab it as:
	                    // sum += 1 and not(item1 xor item2)
	                    // that is, if item1 and item2 are the SAME at bit 1
	                    // then we increase
	                    sum += ( 1 & ((item1 ^ item2) ^ (-1)));
	                    ((GPSemanticIndividual)ind).semantic.semantic[ y ]  =  ( 1L & ((item1 ^ item2) ^(-1L)));
	                    // shift to the next bit
	                    item1 = (byte)(item1 >>> 1);
	                    item2 = (byte)(item2 >>> 1);
	                    }
	                }
	            else if (bits==2)
	                {
	                long item1 = input.dat_6;
	                long item2 = Fast.M_6[Fast.M_6_OUTPUT];
	                for(int y=0;y<MultiplexerData.MULTI_6_BITLENGTH;y++)
	                    {
	                    // if the first bit matches, grab it
	                    sum += ( 1L & ((item1 ^ item2) ^(-1L)));
	                    ((GPSemanticIndividual)ind).semantic.semantic[ y ]  =  ( 1L & ((item1 ^ item2) ^(-1L)));
	                    // shift to the next bit
	                    item1 = item1 >>> 1;
	                    item2 = item2 >>> 1;
	                    }
	                }
	            else // bits==3
	                {
	                long item1, item2;
	                for(int y=0;y<MultiplexerData.MULTI_11_NUM_BITSTRINGS; y++)
	                    {
	                    item1 = input.dat_11[y];
	                    item2 = Fast.M_11[Fast.M_11_OUTPUT][y];
	                    //System.out.println("" + y + " ### " + item1 + " " + item2);
	                    for(int z=0;z<MultiplexerData.MULTI_11_BITLENGTH;z++)
	                        {
	                        // if the first bit matches, grab it
	                        sum += ( 1L & ((item1 ^ item2) ^(-1L)));
	    	                ((GPSemanticIndividual)ind).semantic.semantic[ MultiplexerData.MULTI_11_NUM_BITSTRINGS*y+z ]  =  ( 1L & ((item1 ^ item2) ^(-1L)));
	                        // shift to the next bit
	                        item1 = item1 >>> 1;
	                        item2 = item2 >>> 1;
	                        }
	                    }
	                }
	                
	            // the fitness better be KozaFitness!
	            KozaFitness f = ((GPSemanticIndividual) ind).semantic.regressionFit;
	            if (bits==1)
	                f.setStandardizedFitness(state,(float)(Fast.M_3_SIZE - sum));
	            else if (bits==2)
	                f.setStandardizedFitness(state,(float)(Fast.M_6_SIZE - sum));
	            else // (bits==3)
	                f.setStandardizedFitness(state,(float)(Fast.M_11_SIZE - sum));
	            f.hits = sum;
	            ind.evaluated = true;
	            }

	        }

	
	public void finishEvaluating(final EvolutionState state, final int threadnum) {
    	
    	int n = ((GPSemanticIndividual) state.population.subpops[0].individuals[0]).semantic.semantic.length;    	
    	
    	for (int x = 0; x < n; ++x)
			for (int y = 0; y < n; ++y)
				S[x][y] = 0;
    	
    	for (int i = 0; i < state.population.subpops[0].individuals.length; ++i) {
    		GPSemanticIndividual ind = (GPSemanticIndividual) state.population.subpops[0].individuals[i];
    		for (int x = 0; x < n; ++x)
    			for (int y = 0; y < n; ++y)
    				S[x][y] += ((GPSemantic)(ind.semantic)).semantic[x] * ((GPSemantic)(ind.semantic)).semantic[y];     				
    	}
    	    	    	
    	int m = state.population.subpops[0].individuals.length;
    	for (int i = 0; i < m; ++i) {
    		GPSemanticIndividual ind = (GPSemanticIndividual) state.population.subpops[0].individuals[i];
    		double sum = 0;
    		
    		//prim: fit(I) = \sum_{i,j,S[i,j]!=0} I[i,j] * 1 / S[i,j]    		
    		//zij := si*sj/m - sij
    		//trio: fit(I) = \sum_{i,j,i>j} I[i,j] * 1 / (m/4 - zij)
    		    		    	    	
    		double [] I = ((GPSemantic)(ind.semantic)).semantic;

    		if (type.equals(COSOLVE_IFS))
	    		for (int x = 0; x < n; ++x) {
	    			int y = x;
	    			if (S[x][y] != 0)
	    				if (fit.equals(COSOLVE_PRIM)) {
	    					sum += I[x] * I[y] / S[x][y];
	    				} else {
	    					sum += I[x] * I[y] / (m/4 - S[x][x]*S[y][y]/m + S[x][y]);
	    				}
	    		}
    		
    		if (type.equals(COSOLVE_IFS2))
	    		for (int x = 0; x < n; ++x)
	    			for (int y = x + 1; y < n; ++y)
	    				if (S[x][y] != 0)
	    					if (fit.equals(COSOLVE_PRIM)) {
		    					sum += I[x] * I[y] / S[x][y];
		    				} else {
		    					sum += I[x] * I[y] / (m/4 - S[x][x]*S[y][y]/m + S[x][y]);
		    				}
    		    	
    		SimpleFitness f = ind.semantic.cosolvableFit;    		 
    		f.setFitness( state, (float)sum, false);
    		
    		if (cosolve) {
    			( (SimpleFitness) ind.fitness ).setFitness( state, ind.semantic.cosolvableFit.fitness(), false);
    		} else {
    			( (SimpleFitness) ind.fitness ).setFitness( state, ind.semantic.regressionFit.fitness(), false);
    		}
    		
    		ind.evaluated = true;    		
    	}
    }


	
	
}
