package ec.gp.semantic;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.io.LineNumberReader;

import ec.EvolutionState;
import ec.Prototype;
import ec.gp.GPDefaults;
import ec.gp.koza.KozaFitness;
import ec.simple.SimpleFitness;
import ec.util.Code;
import ec.util.DecodeReturn;
import ec.util.Parameter;

public class GPSemantic implements Prototype {

	public static final String P_SEMANTIC = "semantic";
	public static final String P_SEMANTIC_SIZE = "size";
	
	public double semantic[] = null;
	public KozaFitness   regressionFit = null;
	public SimpleFitness cosolvableFit = null;
			
	public Parameter defaultBase() {
    	return GPDefaults.base().push(P_SEMANTIC);
	}
	
	public void setup(final EvolutionState state, final Parameter base) {	
		Parameter def = defaultBase();		
		int semanticSize = state.parameters.getInt(base.push(P_SEMANTIC_SIZE),def.push(P_SEMANTIC_SIZE));
        semantic = new double[ semanticSize ];
        regressionFit = new KozaFitness();
        cosolvableFit = new SimpleFitness();
	}

	public Object clone() {
		GPSemantic myobj = new GPSemantic();
		myobj.semantic = new double[ semantic.length ];
		for (int i = 0; i < semantic.length; ++i) 
			myobj.semantic[ i ] = semantic[ i ];
		myobj.regressionFit = (KozaFitness) regressionFit.clone();
		myobj.cosolvableFit = (SimpleFitness) cosolvableFit.clone();
		return myobj;
	}
	
		
	public String semanticToStringForHumans() {
		String s = "";
        for (int i = 0; i < semantic.length; i++)
            s = s + " " + semantic[i];
        return s;
	}
	
	
	public String semanticToString() {		
        StringBuffer s = new StringBuffer();
        s.append(Code.encode(semantic.length));
        for (int i = 0; i < semantic.length; i++)
            s.append(Code.encode(semantic[i]));
        return s.toString();
	}
	
	
	public void parseSemantic(final EvolutionState state, final LineNumberReader reader) throws IOException {
        // read in the next line. The first item is the size semantic
        String s = reader.readLine();
        DecodeReturn d = new DecodeReturn(s);
        Code.decode(d);
        int len = (int) (d.l);
        
        if (semantic == null || semantic.length != len)
 		   semantic = new double[len];

        // read in the semantic
        for (int i = 0; i < semantic.length; i++) {
            Code.decode(d);
            semantic[i] = d.d;
        }    
	}
	
	
	public void writeSemantic(final EvolutionState state, final DataOutput dataOutput) throws IOException {
		dataOutput.writeInt(semantic.length);
		for (int i = 0; i < semantic.length; ++i)
			dataOutput.writeDouble(semantic[ i ]);
    }
	
	
	public void readSemantic(final EvolutionState state, final DataInput dataInput) throws IOException {
	   int len = dataInput.readInt();
	   if (semantic == null || semantic.length != len)
		   semantic = new double[len];
	   for (int i = 0; i < semantic.length; ++i) 
		   semantic[i] = dataInput.readDouble();
    }

}
