package ec.gp.semantic;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.PrintWriter;

import ec.EvolutionState;
import ec.gp.GPIndividual;
import ec.util.Parameter;

public class GPSemanticIndividual extends GPIndividual {
	public static final String P_SEMANTIC_SIZE = "semantic-size";
	public GPSemantic semantic = null;
	
	public void setup(final EvolutionState state, final Parameter base) {
		super.setup(state,base);  // necessary (GPIndividual.setup() is NOT empty)
		// that is all, semantic is set as a clone of GPSemanticSpecies.s_prototype in GPSemanticSpecies.newIndividual()
		// and GPSemanticSpecies.s_prototype is instatiated in GPSemanticSpecies.setup()
	}

	//deep clone - force GPIndividual and GPSemantic to copy itself
	public Object clone() {
		GPSemanticIndividual myobj = (GPSemanticIndividual)(super.clone());
		myobj.semantic = (GPSemantic) semantic.clone();		
		return myobj;
	}
	
	//light clone - force only GPSementic, do not force the tree to copy
	public GPSemanticIndividual lightClone() {
		 GPSemanticIndividual myobj = (GPSemanticIndividual)(super.lightClone());
		 if (semantic != null) myobj.semantic = (GPSemantic) semantic.clone();
		 return myobj;
	}
	
	/** Overridden for the semantic */
    public void printIndividualForHumans(final EvolutionState state, final int log, final int verbosity) {
	    super.printIndividualForHumans(state, log, verbosity);
	    state.output.println( semantic.semanticToStringForHumans(), verbosity, log);
    }

    /** Overridden for the semantic */
	public void printIndividual(final EvolutionState state, final int log, final int verbosity) {
		super.printIndividual(state, log, verbosity);
		state.output.println( semantic.semanticToString(), verbosity, log);	
	}

    /** Overridden for the semantic */
	public void printIndividual(final EvolutionState state, final PrintWriter writer) {
		super.printIndividual(state, writer);
		writer.println( semantic.semanticToString() );
	}

    /** Overridden for the semantic */
	public void readIndividual(final EvolutionState state, final LineNumberReader reader) throws IOException {
		super.readIndividual(state, reader);		
		semantic.parseSemantic(state, reader);
	}

    /** Overridden for the semantic */
	public void writeIndividual(final EvolutionState state, final DataOutput dataOutput) throws IOException {
		super.writeIndividual(state, dataOutput);
		semantic.writeSemantic(state, dataOutput);
	}

    /** Overridden for the semantic */
	public void readIndividual(final EvolutionState state, final DataInput dataInput) throws IOException {
		super.readIndividual(state, dataInput);
		semantic.readSemantic(state,dataInput);
	}
}
