package ec.gp.semantic;


import java.io.DataInput;
import java.io.IOException;
import java.io.LineNumberReader;

import ec.EvolutionState;
import ec.Fitness;
import ec.Individual;
import ec.gp.GPSpecies;
import ec.util.Parameter;

public class GPSemanticSpecies extends GPSpecies {
	/** The prototypical semantic for this species. */
    public GPSemantic s_prototype;
    
    public static final String P_SEMANTIC = "semantic";
    
    // add cloning of semantic prototype
    public Object clone() {
    	GPSemanticSpecies myobj = (GPSemanticSpecies) (super.clone());        
    	myobj.s_prototype = (GPSemantic) s_prototype.clone();
    	return myobj;
    } 
    
    public void setup(final EvolutionState state, final Parameter base) {
    	super.setup(state,base);  // necessary (GPSpeciec.setup() is NOT empty)
        if (!(i_prototype instanceof GPSemanticIndividual))
            state.output.fatal("The Individual class for the Species " + getClass().getName() + " is must be a subclass of ec.gp.semantic.GPSemanticIndividual.", base );
     
	    Parameter def = defaultBase();
	    // set up prototype semantic
	    s_prototype = (GPSemantic)(state.parameters.getInstanceForParameterEq(base.push(P_SEMANTIC), def.push(P_SEMANTIC), GPSemantic.class));	    
	    s_prototype.setup(state, base.push(P_SEMANTIC));
	}
    
    public Individual newIndividual(EvolutionState state, int thread) {
    	GPSemanticIndividual newind = (GPSemanticIndividual) ((GPSemanticIndividual)(i_prototype)).lightClone();
        
        // Initialize the trees
        for (int x=0;x<newind.trees.length;x++)
            newind.trees[x].buildTree(state, thread);

        // Set the fitness
        newind.fitness = (Fitness)(f_prototype.clone());
        newind.evaluated = false;

        // Set the species to me
        newind.species = this;

	    // set the semantic
	    newind.semantic = (GPSemantic)(s_prototype.clone());
	    
	    // ...and we're ready!
	    return newind;
    }


// A custom version of newIndividual() which guarantees that the
// prototype is light-cloned before readIndividual is issued
    public Individual newIndividual(final EvolutionState state, final LineNumberReader reader) throws IOException {
    	GPSemanticIndividual newind = (GPSemanticIndividual) ((GPSemanticIndividual)(i_prototype)).lightClone();
    		            
	    // Set the fitness -- must be done BEFORE loading!
	    newind.fitness = (Fitness)(f_prototype.clone());
	    newind.evaluated = false; // for sanity's sake, though it's a useless line
	
	    // load that sucker
	    newind.readIndividual(state,reader);
	
	    // Set the species to me
	    newind.species = this;
	    
	    // set the semantic
	    newind.semantic = (GPSemantic)(s_prototype.clone());

	    // and we're ready!
	    return newind;  
    }


// A custom version of newIndividual() which guarantees that the
// prototype is light-cloned before readIndividual is issued
public Individual newIndividual(final EvolutionState state,
                                final DataInput dataInput)
    throws IOException
    {
    GPSemanticIndividual newind = (GPSemanticIndividual) ((GPSemanticIndividual)i_prototype).lightClone();
    
    // Set the fitness -- must be done BEFORE loading!
    newind.fitness = (Fitness)(f_prototype.clone());
    newind.evaluated = false; // for sanity's sake, though it's a useless line

    // Set the species to me
    newind.species = this;

    // load that sucker
    newind.readIndividual(state,dataInput);

    // set the semantic
    newind.semantic = (GPSemantic)(s_prototype.clone());
    
    // and we're ready!
    return newind;  
    }

}


	
